@@ -4,21 +4,29 @@ import os
 import os.path as op
 from pathlib import Path
 import random
-import requests
 from shutil import rmtree
 import sys
 import time
 
-from .dandiapi import DandiAPIClient
-from . import girder, get_logger
-from .consts import dandiset_metadata_file, metadata_digests
-from .dandiarchive import navigate_url
-from .dandiset import Dandiset
-from .utils import ensure_datetime, flattened, is_same_time
-
-
 import humanize
+import requests
+
+from . import get_logger
+from .consts import RETRY_STATUSES, dandiset_metadata_file
+from .dandiarchive import DandisetURL, MultiAssetURL, SingleAssetURL, parse_dandi_url
+from .dandiset import Dandiset
+from .support.digests import get_digest
 from .support.pyout import naturalsize
+from .utils import (
+    abbrev_prompt,
+    ensure_datetime,
+    find_files,
+    flattened,
+    is_same_time,
+    on_windows,
+    path_is_subpath,
+    pluralize,
+)
 
 lgr = get_logger()
 
@@ -32,18 +40,36 @@ def download(
     jobs=1,
     get_metadata=True,
     get_assets=True,
+    sync=False,
 ):
     # TODO: unduplicate with upload. For now stole from that one
     # We will again use pyout to provide a neat table summarizing our progress
     # with upload etc
-    import pyout
     from .support import pyout as pyouts
 
+    urls = flattened([urls])
+    if len(urls) > 1:
+        raise NotImplementedError("multiple URLs not supported")
+    if not urls:
+        # if no paths provided etc, we will download dandiset path
+        # we are at, BUT since we are not git -- we do not even know
+        # on which instance it exists!  Thus ATM we would do nothing but crash
+        raise NotImplementedError("No URLs were provided.  Cannot download anything")
+
+    parsed_url = parse_dandi_url(urls[0])
+
+    # TODO: if we are ALREADY in a dandiset - we can validate that it is the
+    # same dandiset and use that dandiset path as the one to download under
+    if isinstance(parsed_url, DandisetURL):
+        output_path = op.join(output_dir, parsed_url.dandiset_id)
+    else:
+        output_path = output_dir
+
     # dandi.cli.formatters are used in cmd_ls to provide switchable
     pyout_style = pyouts.get_style(hide_if_missing=False)
 
     rec_fields = ("path", "size", "done", "done%", "checksum", "status", "message")
-    out = pyout.Tabular(style=pyout_style, columns=rec_fields, max_workers=jobs)
+    out = pyouts.LogSafeTabular(style=pyout_style, columns=rec_fields, max_workers=jobs)
 
     out_helper = PYOUTHelper()
     pyout_style["done"] = pyout_style["size"].copy()
@@ -59,8 +85,8 @@ def download(
         kw["yield_generator_for_fields"] = rec_fields[1:]  # all but path
 
     gen_ = download_generator(
-        urls,
-        output_dir,
+        parsed_url,
+        output_path,
         existing=existing,
         get_metadata=get_metadata,
         get_assets=get_assets,
@@ -84,10 +110,51 @@ def download(
     else:
         raise ValueError(format)
 
+    if sync and not isinstance(parsed_url, SingleAssetURL):
+        with parsed_url.get_client() as client:
+            asset_paths = {asset.path for asset in parsed_url.get_assets(client)}
+        if isinstance(parsed_url, DandisetURL):
+            prefix = os.curdir
+            download_dir = output_path
+        elif isinstance(parsed_url, MultiAssetURL):
+            folder_path = op.normpath(parsed_url.path)
+            prefix = folder_path
+            download_dir = op.join(output_path, op.basename(folder_path))
+        else:
+            raise NotImplementedError(
+                f"Unexpected URL type {type(parsed_url).__name__}"
+            )
+        to_delete = []
+        for p in find_files(".*", download_dir, exclude_datalad=True):
+            if p == op.join(output_path, dandiset_metadata_file):
+                continue
+            a_path = op.normpath(op.join(prefix, op.relpath(p, download_dir)))
+            if on_windows:
+                a_path = a_path.replace("\\", "/")
+            if a_path not in asset_paths:
+                to_delete.append(p)
+        if to_delete:
+            while True:
+                opt = abbrev_prompt(
+                    f"Delete {pluralize(len(to_delete), 'local asset')}?",
+                    "yes",
+                    "no",
+                    "list",
+                )
+                if opt == "list":
+                    for p in to_delete:
+                        print(p)
+                elif opt == "yes":
+                    for p in to_delete:
+                        os.unlink(p)
+                    break
+                else:
+                    break
+
 
 def download_generator(
-    urls,
-    output_dir,
+    parsed_url,
+    output_path,
     *,
     assets_it=None,
     yield_generator_for_fields=None,
@@ -99,7 +166,7 @@ def download_generator(
     (as identified by URL)
 
     This function is a generator which would yield records on ongoing activities.
-    Activites include traversal of the remote resource (DANDI archive), download of
+    Activities include traversal of the remote resource (DANDI archive), download of
     individual assets while yielding records (TODO: schema) while validating their
     checksums "on the fly", etc.
 
@@ -110,35 +177,15 @@ def download_generator(
       summary statistics while already downloading.  TODO: reimplement properly!
 
     """
-    urls = flattened([urls])
-    if len(urls) > 1:
-        raise NotImplementedError("multiple URLs not supported")
-    if not urls:
-        # if no paths provided etc, we will download dandiset path
-        # we are at, BUT since we are not git -- we do not even know
-        # on which instance it exists!  Thus ATM we would do nothing but crash
-        raise NotImplementedError("No URLs were provided.  Cannot download anything")
 
-    with navigate_url(urls[0]) as (client, dandiset, assets):
+    with parsed_url.navigate() as (client, dandiset, assets):
         if assets_it:
             assets_it.gen = assets
             assets = assets_it
 
-        # TODO: if we are ALREADY in a dandiset - we can validate that it is the
-        # same dandiset and use that dandiset path as the one to download under
-        if dandiset:
-            identifier = Dandiset._get_identifier(dandiset)
-            if not identifier:
-                raise ValueError(f"Cannot deduce dandiset identifier from {dandiset}")
-            output_path = op.join(output_dir, identifier)
-            if get_metadata:
-                dandiset_metadata = dandiset.get("metadata", {})
-                for resp in _populate_dandiset_yaml(
-                    output_path, dandiset_metadata, existing == "overwrite"
-                ):
-                    yield dict(path=dandiset_metadata_file, **resp)
-        else:
-            output_path = output_dir
+        if isinstance(parsed_url, DandisetURL) and get_metadata:
+            for resp in _populate_dandiset_yaml(output_path, dandiset, existing):
+                yield dict(path=dandiset_metadata_file, **resp)
 
         # TODO: do analysis of assets for early detection of needed renames etc
         # to avoid any need for late treatment of existing and also for
@@ -148,58 +195,56 @@ def download_generator(
             return
 
         for asset in assets:
-            # unavoidable ugliness since girder and API have different "scopes" for
-            # identifying an asset
-            digests_from_metadata = {
-                d: asset.get("metadata")[d]
-                for d in metadata_digests
-                if d in asset.get("metadata", {})
-            }
-            if isinstance(client, girder.GirderCli):
-                down_args = (asset["girder"]["id"],)
-                digests = digests_from_metadata
-            elif isinstance(client, DandiAPIClient):
-                # At least according to our 0.1.0 state of v1
-                # https://github.com/dandi/dandi-publish/issues/79
-                down_args = (
-                    asset["version"]["dandiset"]["identifier"],
-                    asset["version"]["version"],
-                    asset["uuid"],
-                )
-                if "sha256" not in asset:
-                    lgr.warning("For some reason - there no sha256 in %s", str(asset))
-                    digests = digests_from_metadata
-                else:
-                    digests = {"sha256": asset["sha256"]}
-                    if (
-                        "sha256" in digests_from_metadata
-                        and asset["sha256"] != digests_from_metadata["sha256"]
-                    ):
-                        lgr.warning(
-                            "Metadata seems to be outdated since API returned different "
-                            "sha256 for %(path)s",
-                            asset,
-                        )
+            metadata = asset.get_raw_metadata()
+            d = metadata.get("digest", {})
+            if "dandi:dandi-etag" in d:
+                digests = {"dandi-etag": d["dandi:dandi-etag"]}
             else:
-                raise TypeError(f"Don't know here how to handle {client}")
+                raise RuntimeError(
+                    f"dandi-etag not available for asset. Known digests: {d}"
+                )
+            try:
+                digests["sha256"] = d["dandi:sha2-256"]
+            except KeyError:
+                pass
 
-            path = asset["path"].lstrip("/")  # make into relative path
+            path = asset.path.lstrip("/")  # make into relative path
             path = op.normpath(path)
+            if not isinstance(parsed_url, DandisetURL):
+                if isinstance(parsed_url, MultiAssetURL):
+                    folder_path = op.normpath(parsed_url.path)
+                    path = op.join(
+                        op.basename(folder_path), op.relpath(path, folder_path)
+                    )
+                elif isinstance(parsed_url, SingleAssetURL):
+                    path = op.basename(path)
+                else:
+                    raise NotImplementedError(
+                        f"Unexpected URL type {type(parsed_url).__name__}"
+                    )
             download_path = op.join(output_path, path)
 
-            downloader = client.get_download_file_iter(*down_args)
+            downloader = asset.get_download_file_iter()
+
+            try:
+                mtime = metadata["blobDateModified"]
+            except KeyError:
+                mtime = None
+            if mtime is None:
+                lgr.warning(
+                    "Asset %s is missing blobDateModified metadata field",
+                    asset["path"],
+                )
+                mtime = asset.modified
 
-            # Get size from the metadata, although I guess it could be returned directly
-            # by server while establishing downloader... but it seems that girder itself
-            # does get it from the "file" resource, not really from direct URL.  So I guess
-            # we will just follow. For now we must find it in "attrs"
             _download_generator = _download_file(
                 downloader,
                 download_path,
+                toplevel_path=output_path,
                 # size and modified generally should be there but better to redownload
                 # than to crash
-                size=asset.get("size"),
-                mtime=asset.get("modified"),
+                size=asset.size,
+                mtime=mtime,
                 existing=existing,
                 digests=digests,
             )
@@ -232,14 +277,7 @@ class ItemsSummary:
         if not self.files:
             self.t0 = time.time()
         self.files += 1
-        size = rec.get("size")
-        if size is not None:
-            self.size += size
-        elif rec.get("path", "") == "dandiset.yaml":
-            # again -- so special. TODO: make it a proper file
-            pass
-        else:
-            self.has_unknown_sizes = True
+        self.size += rec.size
         return self
 
 
@@ -297,8 +335,7 @@ class PYOUTHelper:
         return v
 
     def agg_done(self, done_sizes):
-        """Formatter for "DONE" column
-        """
+        """Formatter for "DONE" column"""
         done = sum(done_sizes)
         if self.it.finished and done == 0 and self.items_summary.size == 0:
             # even with 0s everywhere consider it 100%
@@ -335,7 +372,8 @@ def _skip_file(msg):
     return {"status": "skipped", "message": str(msg)}
 
 
-def _populate_dandiset_yaml(dandiset_path, metadata, overwrite):
+def _populate_dandiset_yaml(dandiset_path, dandiset, existing):
+    metadata = dandiset.get_raw_metadata()
     if not metadata:
         lgr.warning(
             "Got completely empty metadata record for dandiset, not producing dandiset.yaml"
@@ -343,23 +381,40 @@ def _populate_dandiset_yaml(dandiset_path, metadata, overwrite):
         return
     dandiset_yaml = op.join(dandiset_path, dandiset_metadata_file)
     yield {"message": "updating"}
-    lgr.debug(f"Updating {dandiset_metadata_file} from obtained dandiset " f"metadata")
-    if op.lexists(dandiset_yaml) and not overwrite:
-        yield _skip_file("already exists")
-        return
-    else:
-        dandiset = Dandiset(dandiset_path, allow_empty=True)
-        dandiset.path_obj.mkdir(exist_ok=True)  # exist_ok in case of parallel race
-        old_metadata = dandiset.metadata
-        dandiset.update_metadata(metadata)
-        yield {
-            "status": "done",
-            "message": "updated" if metadata != old_metadata else "same",
-        }
+    lgr.debug("Updating %s from obtained dandiset metadata", dandiset_metadata_file)
+    mtime = dandiset.modified
+    if op.lexists(dandiset_yaml):
+        if existing == "error":
+            raise FileExistsError(dandiset_yaml)
+        elif existing == "refresh" and op.lexists(
+            op.join(dandiset_path, ".git", "annex")
+        ):
+            raise RuntimeError("Not refreshing path in git annex repository")
+        elif existing == "skip" or (
+            existing == "refresh"
+            and os.lstat(dandiset_yaml).st_mtime >= mtime.timestamp()
+        ):
+            yield _skip_file("already exists")
+            return
+    dandiset = Dandiset(dandiset_path, allow_empty=True)
+    dandiset.path_obj.mkdir(exist_ok=True)  # exist_ok in case of parallel race
+    old_metadata = dandiset.metadata
+    dandiset.update_metadata(metadata)
+    os.utime(dandiset_yaml, (time.time(), mtime.timestamp()))
+    yield {
+        "status": "done",
+        "message": "updated" if metadata != old_metadata else "same",
+    }
 
 
 def _download_file(
-    downloader, path, size=None, mtime=None, existing="error", digests=None
+    downloader,
+    path,
+    toplevel_path,
+    size=None,
+    mtime=None,
+    existing="error",
+    digests=None,
 ):
     """Common logic for downloading a single file
 
@@ -373,8 +428,8 @@ def _download_file(
     Parameters
     ----------
     downloader: callable returning a generator
-      A backend (girder or api) specific fixture for downloading some file into
-      path. It should be a generator yielding downloaded blocks.
+      A backend-specific fixture for downloading some file into path. It should
+      be a generator yielding downloaded blocks.
     size: int, optional
       Target size if known
     digests: dict, optional
@@ -383,6 +438,7 @@ def _download_file(
     """
     if op.lexists(path):
         block = f"File {path!r} already exists"
+        annex_path = op.join(toplevel_path, ".git", "annex")
         if existing == "error":
             raise FileExistsError(block)
         elif existing == "skip":
@@ -390,7 +446,39 @@ def _download_file(
             return
         elif existing == "overwrite":
             pass
+        elif existing == "overwrite-different":
+            realpath = op.realpath(path)
+            key_parts = op.basename(realpath).split("-")
+            if size is not None and os.stat(realpath).st_size != size:
+                lgr.debug(
+                    "Size of %s does not match size on server; redownloading", path
+                )
+            elif (
+                op.lexists(annex_path)
+                and op.islink(path)
+                and path_is_subpath(realpath, op.abspath(annex_path))
+                and key_parts[0] == "SHA256E"
+                and digests
+                and "sha256" in digests
+            ):
+                if key_parts[-1].partition(".")[0] == digests["sha256"]:
+                    yield _skip_file("already exists")
+                    return
+                else:
+                    lgr.debug(
+                        "%s is in git-annex, and hash does not match hash on server; redownloading",
+                        path,
+                    )
+            elif get_digest(path, "dandi-etag") == digests["dandi-etag"]:
+                yield _skip_file("already exists")
+                return
+            else:
+                lgr.debug(
+                    "Etag of %s does not match etag on server; redownloading", path
+                )
         elif existing == "refresh":
+            if op.lexists(annex_path):
+                raise RuntimeError("Not refreshing path in git annex repository")
             if mtime is None:
                 lgr.warning(
                     f"{path!r} - no mtime or ctime in the record, redownloading"
@@ -423,7 +511,12 @@ def _download_file(
         # choose first available for now.
         # TODO: reuse that sorting based on speed
         for algo, digest in digests.items():
-            digester = getattr(hashlib, algo, None)
+            if algo == "dandi-etag":
+                from .core.digests.dandietag import ETagHashlike
+
+                digester = lambda: ETagHashlike(size)  # noqa: E731
+            else:
+                digester = getattr(hashlib, algo, None)
             if digester:
                 break
         if not digester:
@@ -431,6 +524,7 @@ def _download_file(
 
     # TODO: how do we discover the total size????
     # TODO: do not do it in-place, but rather into some "hidden" file
+    resuming = False
     for attempt in range(3):
         try:
             if digester:
@@ -439,6 +533,7 @@ def _download_file(
             # I wonder if we could make writing async with downloader
             with DownloadDirectory(path, digests) as dldir:
                 downloaded = dldir.offset
+                resuming = downloaded > 0
                 if size is not None and downloaded == size:
                     # Exit early when downloaded == size, as making a Range
                     # request in such a case results in a 416 error from S3.
@@ -465,14 +560,13 @@ def _download_file(
                     yield msg
                     dldir.append(block)
             break
-            # both girder and we use HttpError
         except requests.exceptions.HTTPError as exc:
             # TODO: actually we should probably retry only on selected codes, and also
             # respect Retry-After
             if attempt >= 2 or exc.response.status_code not in (
                 400,  # Bad Request, but happened with gider:
                 # https://github.com/dandi/dandi-cli/issues/87
-                503,  # Service Unavailable
+                *RETRY_STATUSES,
             ):
                 lgr.debug("Download failed: %s", exc)
                 yield {"status": "error", "message": str(exc)}
@@ -487,7 +581,7 @@ def _download_file(
             )
             time.sleep(random.random() * 5)
 
-    if downloaded_digest:
+    if downloaded_digest and not resuming:
         downloaded_digest = downloaded_digest.hexdigest()  # we care only about hex
         if digest != downloaded_digest:
             msg = f"{algo}: downloaded {downloaded_digest} != {digest}"
@@ -504,7 +598,6 @@ def _download_file(
             # "message": "no digests were provided"
         }
 
-    # It seems that girder might not care about setting either mtime, so we will do if we know
     # TODO: dissolve attrs and pass specific mtime?
     if mtime:
         yield {"status": "setting mtime"}
